from configparser import NoOptionError, NoSectionError

from vja import VjaError
from vja import config
from vja.apiclient import ApiClient

api_client: ApiClient = None


def get_config():
    return config.get_parser()


def do_login(config=None) -> ApiClient:
    """Create and initialize (including authentication) an API client."""
    if not config:
        config = get_config()

    try:
        api_url = config.get("application", "api_url")
    except (NoSectionError, NoOptionError):
        raise VjaError("Login url not specified in %s.Dying." % config.get_path())

    client = ApiClient(api_url=api_url)

    client.authenticate()

    return client


def get_client() -> ApiClient:
    global api_client
    if api_client:
        return api_client
    try:
        api_client = do_login()
        return api_client
    except VjaError as e:
        print(e)
        exit(1)



__all__ = ['apiclient', 'service', 'config', 'login', 'VjaError']

class VjaError(Exception):
    """Internal error"""
    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __repr__(self):
        return 'VjaError("%s")' % self.msg

    def __str__(self):
        return self.msg



import json
import logging
import os
from functools import wraps
from types import SimpleNamespace

import click
import requests

from vja import VjaError, config

logger = logging.getLogger(__name__)

__all__ = ['ApiClient']

def check_access_token(func):
    """ A decorator that makes the decorated function check for access token."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        """Provide decorated function with `access_token` provided as a keword
        argument.
        """
        # check if `access_token` is set to a value in kwargs
        if 'access_token' in kwargs and kwargs['access_token'] is not None:
            return func(*args, **kwargs)
        else:
            # try to get the access token
            try:
                self = args[0]
                self.get_access_token()
                kwargs['access_token'] = self.access_token
                return func(*args, **kwargs)
            # no access_token in kwargs or in class; die
            except KeyError:
                raise VjaError('need access token to call function %s;'
                                    ' call authenticate()' % func.__name__)
    return wrapper


def handle_http_error(func):
    """A decorator to handle some HTTP errors raised in decorated function f."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        """Handle the HTTPError exceptions raised in f."""
        try:
            return func(*args, **kwargs)
        except requests.HTTPError as error:
            if error.response.status_code == 401 or error.response.status_code == 429:
                logger.info("HTTP-Error %s, url=%s; trying to retrieve new access token..."
                            % (error.response.status_code, error.response.url))
                self = args[0]
                self.get_access_token(force=True)
                kwargs['access_token'] = self._token['access']
                return func(*args, **kwargs)  # try again with new token. Re-raise if failed.
            else:
                body = error.response.text
                logger.warning("HTTP-Error %s, url=%s body=%s"
                               % (error.response.status_code, error.response.url, body))
                raise

    return wrapper


class ApiClient(object):
    """Vikunja API client."""
    _TOKEN_FILE = 'token.json'

    def __init__(self, api_url):
        """Initialize a new ApiClient w/o auth credentials."""
        # config
        self._config = dict()
        self._config['application'] = dict()
        self._config['application']['api_url'] = api_url
        # OAuth2 tokens and scope
        self._userid = None
        self._token = dict()
        self._token['access'] = None
        # caches
        self._cache = dict()
        self._cache['lists'] = None
        self._cache['labels'] = None
        self._cache['namespaces'] = None
        self._cache['tasks'] = None

    @property
    def userid(self):
        """Property for accessing the cached userid."""
        if self._userid is None:
            raise KeyError('userid not set! call authenticate()')
        return self._userid

    @property
    def access_token(self):
        """Property for accessing the cached access token."""
        if not self._token['access']:
            raise KeyError('access token not set! call authenticate()')
        return self._token['access']

    def create_url(self, path):
        return self._config['application']['api_url'] + path

    ###
    # Authentication
    ###
    def authenticate(self):
        """Use credentials to get userid and access token."""
        self.get_access_token()
        self._userid = self._userid if self._userid else self.getUserid()

    def getUserid(self):
        userid = self.getAccountInfo().id
        if userid == '1':
            raise VjaError('invalid username/password')
        return userid

    @check_access_token
    def getAccountInfo(self, access_token=None):
        """Retrieve account info (like pro, timezone, and lastedit_task)."""
        url=self.create_url("/user")
        return self.getJson(url)

    def load_access_token(self):
        """Load the access token from the file."""
        token_path = os.path.join(config.get_dir(), ApiClient._TOKEN_FILE)
        try:
            with open(token_path) as token_file:
                data = json.load(token_file)
        except (IOError):
            return False
        self._token['access'] = data['token']
        if not self._token['access']:
            return False
        return True

    def store_access_token(self):
        """Store the access token to the file."""
        token_path = os.path.join(config.get_dir(), ApiClient._TOKEN_FILE)
        data = {'token': self.access_token}
        with open(token_path, 'w') as token_file:
            json.dump(data, token_file)


    @handle_http_error
    def get_access_token(self, force=False):
        """Get access token as specified in the API v3 docs."""
        if self.load_access_token() and not force:
            return
        url = self.create_url("/login")
        username = click.prompt("username")
        password = click.prompt("password", hide_input=True)
        payload = {'username': username,
                   'password': password}
        response = requests.post(url, json=payload)
        response.raise_for_status()
        self._token['access'] = response.json()['token']
        logger.info("Login succesful.")
        self.store_access_token()

    @handle_http_error
    def getJson(self, url, params=None, **kwargs):
        headers = {'Authorization': "Bearer {}".format(self.access_token)}
        response = requests.get(url, headers=headers, params=params)
        response.raise_for_status()
        return response.json(object_hook=lambda d: SimpleNamespace(**d))

    @check_access_token
    def getTasks(self, exclude_completed=True, **kwargs):
        if self._cache['tasks']:
            return self._cache['tasks']
        else:
            url=self.create_url("/tasks/all")
            params={'filter_by': 'done', 'filter_value': 'false'} if exclude_completed else None
            self._cache['tasks'] = self.getJson(url, params)
            return self._cache['tasks']

    @check_access_token
    def getLists(self, **kwargs):
        if self._cache['lists']:
            return self._cache['lists']
        else:
            url=self.create_url("/lists")
            self._cache['lists'] = self.getJson(url)
            return self._cache['lists']

    @check_access_token
    def getLabels(self, **kwargs):
        if self._cache['labels']:
            return self._cache['labels']
        else:
            url=self.create_url("/labels")
            self._cache['labels'] = self.getJson(url)
            return self._cache['labels']



import logging
from collections import defaultdict
from datetime import datetime, timedelta
from time import mktime

import parsedatetime as pdt
from dateutil import tz

from vja.login import get_client
from vja.model import PrintableTask

logger = logging.getLogger(__name__)

def list_tasks(list, list_size, urgency_sort):
    tasks = get_tasks(None, 'today', list)
    if list_size == 0:
        tasks += get_tasks('today', 'in 3 days', list)
        tasks += get_tasks(None, None, list, 'next action')
    if list_size > 0:
        tasks += get_tasks('today', 'tomorrow+364 days', list)
        tasks += get_tasks(None, None, list)
    if (urgency_sort):
        tasks.sort(key=lambda x: x.urgency(), reverse=True)
    print_tasks(tasks, urgency_sort)


def get_tasks(day_start=None, day_end='today', list=None, status=None):
    start = get_max_time(day_start) if day_start else None
    end = get_max_time(day_end) if day_end else None

    raw_tasks = get_client().getTasks(exclude_completed=True)

    filtered_tasks = [PrintableTask(x) for x in raw_tasks if is_in(x, start, end)]
    filtered_tasks.sort(key=lambda x: (x.due_date or datetime.max,
               -(x.priority),
               x.listname.upper(),
               x.title.upper()))
    return filtered_tasks

def get_max_time(day='today'):
    timest = pdt.Calendar().parse(day)[0]
    now = datetime.fromtimestamp(mktime(timest), tz.tzlocal())
    daystart = datetime(now.year, now.month, now.day, tzinfo=now.tzinfo)
    dayend = daystart + timedelta(days=1)
    utc_dayend = dayend.isoformat()
    return utc_dayend

def is_in(item, start_date, end_date):
    if ((not item.due_date and not start_date and not end_date)
        or (item.due_date and end_date and item.due_date < end_date)
        and (not start_date or (item.due_date and item.due_date > start_date))):
        return True
    else:
        return False

def print_tasks(tasks, priority_level_sort=False):
    if not tasks:
        print("No tasks found. Go home early!")
    if (priority_level_sort):
        tasks_by_prio = defaultdict(list)
        for task in tasks:
            tasks_by_prio[task.urgency()].append(task)
        for prio, items in tasks_by_prio.items():
            #print("--------------------------------------")
            print()
            print_task_list(tasks, items)
    else:
        print_task_list(tasks, tasks)

def print_task_list(tasks, items):
    for item in items:
        print(f"{str(tasks.index(item) + 1):2}" + " " + item.representation())



import configparser
import os

from vja import VjaError

CONFIGDIR = os.path.expanduser("~/.vjacli")
CONFIGFILE = os.path.join(CONFIGDIR, "vja.rc")


def get_path():
    """Return default config file path."""
    return CONFIGFILE


def get_dir():
    """Return default config file path."""
    return CONFIGDIR


def load():
    """Load config file."""
    filepath = get_path()
    parser = configparser.SafeConfigParser()
    parser.read(filepath)
    return parser


def store(config):
    """Store the config to the file."""
    if not os.path.exists(get_dir()):
        os.mkdir(get_dir())
    with open(get_path(), "w") as cfile:
        config.write(cfile)


def reload(parser):
    """Reload the config file."""
    filepath = get_path()
    parser.read(filepath)


__parser__ = load()


def get_parser():
    """Get the config parser."""
    if not __parser__:
        raise VjaError("config file is not loaded.")
    else:
        return __parser__



#!/usr/bin/env python

import logging
import webbrowser
from importlib import metadata

import click
from click_aliases import ClickAliasedGroup

from vja import config, service

logger = logging.getLogger(__name__)

@click.group(cls=ClickAliasedGroup)
@click.option('-v', '--verbose', 'verbose', default=False, is_flag=True, help='verbose output')
def cli(verbose):
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
        logger.setLevel(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
        logger.setLevel(level=logging.INFO)
    logger.debug(metadata.version("vja"))


@cli.command(aliases=['list'], help='list relevant tasks (optionaly with -a)')
@click.option('-l', '--list', 'list')
@click.option('-a', '--all', 'list_size', count=True, help='show additional tasks (tomorrow + task without due date)')
@click.option('-i', '--urgency', 'urgency_sort', is_flag=True, help='sort by priority level')
def ls(list=None, list_size=0, urgency_sort=False):
    service.list_tasks(list, list_size, urgency_sort)
    return

@cli.command(help='list new tasks')
def inbox(context=None, list_size=0):
    return


@cli.command(help='show task details')
@click.argument('task', required=True, type=click.INT)
def show(task):
    logger.debug("task: %s" % task.dict())
    print(task)

@cli.command(name='open', help='open task in browser')
@click.argument('task', required=False, type=click.INT)
def open_task(task):
    # TODO restore cache logic
    url = config.get_parser().get('application', 'frontend_url')
    if (task and task > 0):
        url += "/tasks/" + str(task)
    webbrowser.open_new_tab(url)


@cli.command(help='add new task, Example: \'add task do something .saturday @@context\'')
@click.argument('line', nargs=-1)
def add(line):
    return

@cli.command()
@click.argument('task', required=True, type=click.INT)
@click.argument('line', nargs=-1)
# @click.argument('modification', required=True, type=click.STRING)
def edit(task, line):
    return

@cli.command(help='mark task as complete')
@click.argument('task', required=True, type=click.INT)
def check(task):
    return

if __name__ == '__main__':
    cli()



from datetime import datetime

from vja.login import get_client


def format_date(timestamp):
    return timestamp.strftime("%a %d.%m") if timestamp else ""


def format_time(timestamp):
    return timestamp.strftime("%H:%M") if timestamp else ""


class PrintableTask(object):
    def __init__(self, task):
        self.id = task.id
        self.title = task.title
        self.is_favorite = task.is_favorite
        self.priority = task.priority

        self.due_date = None
        if task.due_date and task.due_date > "0001-01-01T00:00:00Z":
            self.due_date = datetime.fromisoformat(
                task.due_date.replace("Z", "")
            ).replace(tzinfo=None)

        lists = get_client().getLists()
        listDict = {x.id: x.title for x in lists}
        self.listname = str(listDict.get(int(task.list_id)))

        self.labelnames = []
        if task.labels:
            labels = get_client().getLabels()
            labelDict = {x.id: x.title for x in labels}
            self.labelnames = [str(labelDict.get(int(x.id))) for x in task.labels]

    def urgency(self):
        today = datetime.today()
        if (self.due_date):
            duedate = self.due_date
            datediff = (duedate - today).days
            if (datediff < 0):
                datepoints = 6
            elif (datediff == 0):
                datepoints = 5
            elif (datediff == 1):
                datepoints = 4
            elif (datediff > 1 and datediff <= 2):
                datepoints = 3
            elif (datediff > 2 and datediff <= 5):
                datepoints = 3
            elif (datediff > 5 and datediff <= 10):
                datepoints = 1
            elif (datediff > 10):
                datepoints = -1
            else:
                datepoints = 0
        else:
            datepoints = 0
        statuspoints = 0
        if ("next" in self.listname.lower() or "next" in " ".join(self.labelnames).lower()):
            statuspoints = 1
        return 2 + statuspoints + datepoints + int(self.priority) + (1 if self.is_favorite else 0)


    def due_date_format(self):
        return format_date(self.due_date)

    def due_time_format(self):
        return format_time(self.due_date)

    def label_names(self):
        return " ".join(self.labelnames)

    def representation(self):
        output = ["(%s)" % self.priority,
                  "%s" % "*" if self.is_favorite else " ",
                  f"{self.title:50.50}",
                  f"{self.due_date_format():9.9}",
                  f"{self.due_time_format():5.5}",
                  f"{self.listname:15.15}",
                  f"{self.label_names():15.15}",
                  f"{self.urgency():3}"]
        return " ".join(output)
